@@ -23,6 +23,14 @@ struct randr_mode {
 	bool preferred;
 };
 
+enum randr_head_prop {
+	RANDR_HEAD_MODE = 1 << 0,
+	RANDR_HEAD_POSITION = 1 << 1,
+	RANDR_HEAD_TRANSFORM = 1 << 2,
+	RANDR_HEAD_SCALE = 1 << 3,
+	RANDR_HEAD_ADAPTIVE_SYNC = 1 << 4,
+};
+
 struct randr_head {
 	struct randr_state *state;
 	struct zwlr_output_head_v1 *wlr_head;
@@ -33,6 +41,7 @@ struct randr_head {
 	int32_t phys_width, phys_height; // mm
 	struct wl_list modes;
 
+	uint32_t changed; // enum randr_head_prop
 	bool enabled;
 	struct randr_mode *mode;
 	struct {
@@ -50,6 +59,7 @@ struct randr_state {
 
 	struct wl_list heads;
 	uint32_t serial;
+	bool has_serial;
 	bool running;
 	bool failed;
 };
@@ -132,6 +142,153 @@ static void print_state(struct randr_state *state) {
 	state->running = false;
 }
 
+static void print_json_optional_string(const char *str) {
+	if (str == NULL) {
+		printf("null");
+		return;
+	}
+
+	printf("\"");
+	for (size_t i = 0; str[i] != '\0'; i++) {
+		char ch = str[i];
+		switch (ch) {
+		case '"':
+			printf("\\\"");
+			break;
+		case '\\':
+			printf("\\\\");
+			break;
+		case '\b':
+			printf("\\b");
+			break;
+		case '\f':
+			printf("\\f");
+			break;
+		case '\n':
+			printf("\\n");
+			break;
+		case '\r':
+			printf("\\r");
+			break;
+		case '\t':
+			printf("\\t");
+			break;
+		default:
+			if (ch > 0 && ch < 0x20) {
+				printf("\\u%04x", ch);
+			} else {
+				printf("%c", ch);
+			}
+		}
+	}
+	printf("\"");
+}
+
+static void print_state_json(struct randr_state *state) {
+	uint32_t version = zwlr_output_manager_v1_get_version(state->output_manager);
+
+	printf("[");
+
+	size_t heads_count = 0;
+	struct randr_head *head;
+	wl_list_for_each(head, &state->heads, link) {
+		if (heads_count++) {
+			printf(",");
+		}
+		printf("\n  {\n");
+
+		printf("    \"name\": ");
+		print_json_optional_string(head->name);
+		printf(",\n");
+
+		printf("    \"description\": ");
+		print_json_optional_string(head->description);
+		printf(",\n");
+
+		printf("    \"make\": ");
+		print_json_optional_string(head->make);
+		printf(",\n");
+
+		printf("    \"model\": ");
+		print_json_optional_string(head->model);
+		printf(",\n");
+
+		printf("    \"serial\": ");
+		print_json_optional_string(head->serial_number);
+		printf(",\n");
+
+		printf("    \"physical_size\": {\n");
+		printf("      \"width\": %d,\n", head->phys_width);
+		printf("      \"height\": %d\n", head->phys_height);
+		printf("    },\n");
+
+		printf("    \"enabled\": %s,\n", head->enabled ? "true" : "false");
+
+		printf("    \"modes\": [");
+
+		size_t modes_count = 0;
+		struct randr_mode *mode;
+		wl_list_for_each(mode, &head->modes, link) {
+			if (modes_count++) {
+				printf(",");
+			}
+			printf("\n      {\n");
+
+			printf("        \"width\": %d,\n", mode->width);
+			printf("        \"height\": %d,\n", mode->height);
+			printf("        \"refresh\": %f,\n", (float)mode->refresh / 1000);
+			printf("        \"preferred\": %s,\n", mode->preferred ? "true" : "false");
+			printf("        \"current\": %s\n", head->mode == mode ? "true" : "false");
+
+			printf("      }");
+		}
+
+		if (modes_count) {
+			printf("\n    ");
+		}
+		printf("]");
+
+		if (!head->enabled) {
+			printf("\n");
+		} else {
+			printf(",\n");
+
+			printf("    \"position\": {\n");
+			printf("      \"x\": %d,\n", head->x);
+			printf("      \"y\": %d\n", head->y);
+			printf("    },\n");
+
+			printf("    \"transform\": ");
+			print_json_optional_string(output_transform_map[head->transform]);
+			printf(",\n");
+
+			printf("    \"scale\": %f,\n", head->scale);
+
+			char *adaptive_sync = "null";
+			if (version >= 4) {
+				switch (head->adaptive_sync_state) {
+				case ZWLR_OUTPUT_HEAD_V1_ADAPTIVE_SYNC_STATE_ENABLED:
+					adaptive_sync = "true";
+					break;
+				case ZWLR_OUTPUT_HEAD_V1_ADAPTIVE_SYNC_STATE_DISABLED:
+					adaptive_sync = "false";
+					break;
+				}
+			}
+			printf("    \"adaptive_sync\": %s\n", adaptive_sync);
+		}
+
+		printf("  }");
+	}
+
+	if (heads_count) {
+		printf("\n");
+	}
+	printf("]\n");
+
+	state->running = false;
+}
+
 static void config_handle_succeeded(void *data,
 		struct zwlr_output_configuration_v1 *config) {
 	struct randr_state *state = data;
@@ -180,25 +337,35 @@ static void apply_state(struct randr_state *state, bool dry_run) {
 
 		struct zwlr_output_configuration_head_v1 *config_head =
 			zwlr_output_configuration_v1_enable_head(config, head->wlr_head);
-		if (head->mode != NULL) {
-			zwlr_output_configuration_head_v1_set_mode(config_head,
-				head->mode->wlr_mode);
-		} else {
-			zwlr_output_configuration_head_v1_set_custom_mode(config_head,
-				head->custom_mode.width, head->custom_mode.height,
-				head->custom_mode.refresh);
+		if (head->changed & RANDR_HEAD_MODE) {
+			if (head->mode != NULL) {
+				zwlr_output_configuration_head_v1_set_mode(config_head,
+					head->mode->wlr_mode);
+			} else {
+				zwlr_output_configuration_head_v1_set_custom_mode(config_head,
+					head->custom_mode.width, head->custom_mode.height,
+					head->custom_mode.refresh);
+			}
 		}
-		zwlr_output_configuration_head_v1_set_position(config_head,
-			head->x, head->y);
-		zwlr_output_configuration_head_v1_set_transform(config_head,
-			head->transform);
-		zwlr_output_configuration_head_v1_set_scale(config_head,
-			wl_fixed_from_double(head->scale));
-
-		if (zwlr_output_manager_v1_get_version(state->output_manager) >= 4) {
+		if (head->changed & RANDR_HEAD_POSITION) {
+			zwlr_output_configuration_head_v1_set_position(config_head,
+				head->x, head->y);
+		}
+		if (head->changed & RANDR_HEAD_TRANSFORM) {
+			zwlr_output_configuration_head_v1_set_transform(config_head,
+				head->transform);
+		}
+		if (head->changed & RANDR_HEAD_SCALE) {
+			zwlr_output_configuration_head_v1_set_scale(config_head,
+				wl_fixed_from_double(head->scale));
+		}
+		if (head->changed & RANDR_HEAD_ADAPTIVE_SYNC) {
+			assert(zwlr_output_manager_v1_get_version(state->output_manager) >=
+				ZWLR_OUTPUT_CONFIGURATION_HEAD_V1_SET_ADAPTIVE_SYNC_SINCE_VERSION);
 			zwlr_output_configuration_head_v1_set_adaptive_sync(config_head,
 				head->adaptive_sync_state);
 		}
+		zwlr_output_configuration_head_v1_destroy(config_head);
 	}
 
 	if (dry_run) {
@@ -395,6 +562,7 @@ static void output_manager_handle_done(void *data,
 		struct zwlr_output_manager_v1 *manager, uint32_t serial) {
 	struct randr_state *state = data;
 	state->serial = serial;
+	state->has_serial = true;
 }
 
 static void output_manager_handle_finished(void *data,
@@ -434,6 +602,7 @@ static const struct wl_registry_listener registry_listener = {
 static const struct option long_options[] = {
 	{"help", no_argument, 0, 'h'},
 	{"dryrun", no_argument, 0, 0},
+	{"json", no_argument, 0, 0},
 	{"output", required_argument, 0, 0},
 	{"on", no_argument, 0, 0},
 	{"off", no_argument, 0, 0},
@@ -560,6 +729,7 @@ static bool parse_output_arg(struct randr_head *head,
 			return false;
 		}
 
+		head->changed |= RANDR_HEAD_MODE;
 		head->mode = mode;
 		head->custom_mode.width = 0;
 		head->custom_mode.height = 0;
@@ -579,6 +749,7 @@ static bool parse_output_arg(struct randr_head *head,
 			return false;
 		}
 
+		head->changed |= RANDR_HEAD_MODE;
 		head->mode = mode;
 		head->custom_mode.width = 0;
 		head->custom_mode.height = 0;
@@ -589,6 +760,7 @@ static bool parse_output_arg(struct randr_head *head,
 			return false;
 		}
 
+		head->changed |= RANDR_HEAD_MODE;
 		head->mode = NULL;
 		head->custom_mode.width = width;
 		head->custom_mode.height = height;
@@ -609,6 +781,7 @@ static bool parse_output_arg(struct randr_head *head,
 			return false;
 		}
 
+		head->changed |= RANDR_HEAD_POSITION;
 		head->x = x;
 		head->y = y;
 	} else if (strcmp(name, "transform") == 0) {
@@ -627,6 +800,8 @@ static bool parse_output_arg(struct randr_head *head,
 			fprintf(stderr, "invalid transform: %s\n", value);
 			return false;
 		}
+
+		head->changed |= RANDR_HEAD_TRANSFORM;
 	} else if (strcmp(name, "scale") == 0) {
 		char *end;
 		double scale = strtod(value, &end);
@@ -635,9 +810,11 @@ static bool parse_output_arg(struct randr_head *head,
 			return false;
 		}
 
+		head->changed |= RANDR_HEAD_SCALE;
 		head->scale = scale;
 	} else if (strcmp(name, "adaptive-sync") == 0) {
-		if (zwlr_output_head_v1_get_version(head->wlr_head) < 4) {
+		if (zwlr_output_head_v1_get_version(head->wlr_head) <
+				ZWLR_OUTPUT_CONFIGURATION_HEAD_V1_SET_ADAPTIVE_SYNC_SINCE_VERSION) {
 			fprintf(stderr, "setting adaptive sync not supported by the compositor\n");
 			return false;
 		}
@@ -649,6 +826,7 @@ static bool parse_output_arg(struct randr_head *head,
 			fprintf(stderr, "invalid adaptive sync state: %s\n", value);
 			return false;
 		}
+		head->changed |= RANDR_HEAD_ADAPTIVE_SYNC;
 	} else {
 		fprintf(stderr, "invalid option: %s\n", name);
 		return false;
@@ -661,6 +839,7 @@ static const char usage[] =
 	"usage: wlr-randr [options…]\n"
 	"--help\n"
 	"--dryrun\n"
+	"--json\n"
 	"--output <name>\n"
 	"  --on\n"
 	"  --off\n"
@@ -684,8 +863,11 @@ int main(int argc, char *argv[]) {
 
 	struct wl_registry *registry = wl_display_get_registry(display);
 	wl_registry_add_listener(registry, &registry_listener, &state);
-	wl_display_dispatch(display);
-	wl_display_roundtrip(display);
+
+	if (wl_display_roundtrip(display) < 0) {
+		fprintf(stderr, "wl_display_roundtrip failed\n");
+		return EXIT_FAILURE;
+	}
 
 	if (state.output_manager == NULL) {
 		fprintf(stderr, "compositor doesn't support "
@@ -693,14 +875,14 @@ int main(int argc, char *argv[]) {
 		return EXIT_FAILURE;
 	}
 
-	while (state.serial == 0) {
+	while (!state.has_serial) {
 		if (wl_display_dispatch(display) < 0) {
 			fprintf(stderr, "wl_display_dispatch failed\n");
 			return EXIT_FAILURE;
 		}
 	}
 
-	bool changed = false, dry_run = false;
+	bool changed = false, dry_run = false, json = false;
 	struct randr_head *current_head = NULL;
 	while (1) {
 		int option_index = -1;
@@ -730,6 +912,8 @@ int main(int argc, char *argv[]) {
 			}
 		} else if (strcmp(name, "dryrun") == 0) {
 			dry_run = true;
+		} else if (strcmp(name, "json") == 0) {
+			json = true;
 		} else { // output sub-option
 			if (current_head == NULL) {
 				fprintf(stderr, "no --output specified before --%s\n", name);
@@ -746,6 +930,8 @@ int main(int argc, char *argv[]) {
 
 	if (changed) {
 		apply_state(&state, dry_run);
+	} else if (json) {
+		print_state_json(&state);
 	} else {
 		print_state(&state);
 	}
@@ -754,7 +940,21 @@ int main(int argc, char *argv[]) {
 		// This space intentionally left blank
 	}
 
-	// TODO: destroy heads
+	struct randr_head *head, *tmp_head;
+	wl_list_for_each_safe(head, tmp_head, &state.heads, link) {
+		struct randr_mode *mode, *tmp_mode;
+		wl_list_for_each_safe(mode, tmp_mode, &head->modes, link) {
+			zwlr_output_mode_v1_destroy(mode->wlr_mode);
+			free(mode);
+		}
+		zwlr_output_head_v1_destroy(head->wlr_head);
+		free(head->name);
+		free(head->description);
+		free(head->make);
+		free(head->model);
+		free(head->serial_number);
+		free(head);
+	}
 	zwlr_output_manager_v1_destroy(state.output_manager);
 	wl_registry_destroy(registry);
 	wl_display_disconnect(display);
